@@ -17,14 +17,14 @@ from requests_oauthlib import OAuth1
 import mwclient.errors as errors
 import mwclient.listing as listing
 from mwclient.sleep import Sleepers
-from mwclient.util import parse_timestamp
+from mwclient.util import parse_timestamp, read_in_chunks
 
 try:
     import gzip
 except ImportError:
     gzip = None
 
-__ver__ = '0.8.5'
+__ver__ = '0.9.3'
 
 log = logging.getLogger(__name__)
 
@@ -63,6 +63,8 @@ class Site(object):
         self.max_lag = text_type(max_lag)
         self.force_login = force_login
         self.requests = reqs or {}
+        if 'timeout' not in self.requests:
+            self.requests['timeout'] = 30  # seconds
 
         if consumer_token is not None:
             auth = OAuth1(consumer_token, consumer_secret, access_token, access_secret)
@@ -119,6 +121,9 @@ class Site(object):
         # Initialization status
         self.initialized = False
 
+        # Upload chunk size in bytes
+        self.chunk_size = 1048576
+
         if do_init:
             try:
                 self.site_init()
@@ -307,6 +312,13 @@ class Site(object):
                 sleeper.sleep()
                 return False
 
+            # cope with https://phabricator.wikimedia.org/T106066
+            if (info['error'].get('code') == u'mwoauth-invalid-authorization' and
+               'Nonce already used' in info['error'].get('info')):
+                log.warning('retrying due to nonce error https://phabricator.wikimedia.org/T106066')
+                sleeper.sleep()
+                return False
+
             if 'query' in info['error']:
                 # Semantic Mediawiki does not follow the standard error format
                 raise errors.APIError(None, info['error']['query'], kwargs)
@@ -342,6 +354,7 @@ class Site(object):
             data (dict): Post data
             files (dict): Files to upload
             retry_on_error (bool): Retry on connection error
+            http_method (str): The HTTP method, defaults to 'POST'
 
         Returns:
             The raw text response.
@@ -362,12 +375,16 @@ class Site(object):
 
         while True:
             try:
+                args = {'files': files, 'headers': headers}
+                for k, v in self.requests.items():
+                    args[k] = v
                 if http_method == 'GET':
-                    stream = self.connection.get(url, params=data, files=files,
-                                                 headers=headers, **self.requests)
+                    args['params'] = data
                 else:
-                    stream = self.connection.post(url, data=data, files=files,
-                                                  headers=headers, **self.requests)
+                    args['data'] = data
+
+                stream = self.connection.request(http_method, url, **args)
+
                 if stream.headers.get('x-database-lag'):
                     wait_time = int(stream.headers.get('retry-after'))
                     log.warning('Database lag exceeds max lag. '
@@ -408,7 +425,7 @@ class Site(object):
                             http_method=http_method)
 
         try:
-            return json.loads(res)
+            return json.loads(res, object_pairs_hook=OrderedDict)
         except ValueError:
             if res.startswith('MediaWiki API is not enabled for this site.'):
                 raise errors.APIDisabledError
@@ -500,7 +517,7 @@ class Site(object):
             # have to try. If the attempt fails, we try the old method.
             try:
                 kwargs['lgtoken'] = self.get_token('login')
-            except KeyError:
+            except (errors.APIError, KeyError):
                 log.debug('Failed to get login token, MediaWiki is older than 1.27.')
 
             while True:
@@ -605,20 +622,34 @@ class Site(object):
         if not image.can('upload'):
             raise errors.InsufficientPermission(filename)
 
-        predata = {}
-
         if comment is None:
-            predata['comment'] = description
+            comment = description
+            text = None
         else:
-            predata['comment'] = comment
-            predata['text'] = description
+            comment = comment
+            text = description
+
+        if file is not None:
+            if not hasattr(file, 'read'):
+                file = open(file, 'rb')
+
+            content_size = file.seek(0, 2)
+            file.seek(0)
+
+            if self.version[:2] >= (1, 20) and content_size > self.chunk_size:
+                return self.chunk_upload(file, filename, ignore, comment, text)
+
+        predata = {
+            'action': 'upload',
+            'format': 'json',
+            'filename': filename,
+            'comment': comment,
+            'text': text,
+            'token': image.get_token('edit'),
+        }
 
         if ignore:
             predata['ignorewarnings'] = 'true'
-        predata['token'] = image.get_token('edit')
-        predata['action'] = 'upload'
-        predata['format'] = 'json'
-        predata['filename'] = filename
         if url:
             predata['url'] = url
 
@@ -638,11 +669,7 @@ class Site(object):
             # Since the filename in Content-Disposition is not interpreted,
             # we can send some ascii-only dummy name rather than the real
             # filename, which might contain non-ascii.
-            file = ('fake-filename', file)
-            # End of workaround
-            # ----------------------------------------------------------------
-
-            files = {'file': file}
+            files = {'file': ('fake-filename', file)}
 
         sleeper = self.sleepers.make()
         while True:
@@ -651,7 +678,72 @@ class Site(object):
             if not info:
                 info = {}
             if self.handle_api_result(info, kwargs=predata, sleeper=sleeper):
-                return info.get('upload', {})
+                response = info.get('upload', {})
+                break
+        if file is not None:
+            file.close()
+        return response
+
+    def chunk_upload(self, file, filename, ignorewarnings, comment, text):
+        """Upload a file to the site in chunks.
+
+        This method is called by `Site.upload` if you are connecting to a newer
+        MediaWiki installation, so it's normally not necessary to call this
+        method directly.
+
+        Args:
+            file (file-like object): File object or stream to upload.
+            params (dict): Dict containing upload parameters.
+        """
+        image = self.Images[filename]
+
+        content_size = file.seek(0, 2)
+        file.seek(0)
+
+        params = {
+            'action': 'upload',
+            'format': 'json',
+            'stash': 1,
+            'offset': 0,
+            'filename': filename,
+            'filesize': content_size,
+            'token': image.get_token('edit'),
+        }
+        if ignorewarnings:
+            params['ignorewarnings'] = 'true'
+
+        sleeper = self.sleepers.make()
+        offset = 0
+        for chunk in read_in_chunks(file, self.chunk_size):
+            while True:
+                data = self.raw_call('api', params, files={'chunk': chunk})
+                info = json.loads(data)
+                if self.handle_api_result(info, kwargs=params, sleeper=sleeper):
+                    response = info.get('upload', {})
+                    break
+
+            offset += chunk.tell()
+            chunk.close()
+            log.debug('%s: Uploaded %d of %d bytes', filename, offset, content_size)
+            params['filekey'] = response['filekey']
+            if response['result'] == 'Continue':
+                params['offset'] = response['offset']
+            elif response['result'] == 'Success':
+                file.close()
+                break
+            else:
+                # Some kind or error or warning occured. In any case, we do not
+                # get the parameters we need to continue, so we should return
+                # the response now.
+                file.close()
+                return response
+
+        del params['action']
+        del params['stash']
+        del params['offset']
+        params['comment'] = comment
+        params['text'] = text
+        return self.post('upload', **params)
 
     def parse(self, text=None, title=None, page=None, prop=None,
               redirects=False, mobileformat=False):
